"""Tests for feed module."""

from accern import StreamClient, StreamListener, error
# from mock import Mock
import json
import pytest

api_base = "http://http://feed-staging.accern.com/v4/stream"


class MyStreamListener(StreamListener):
    def on_data(self, data):
        data_json = json.loads(data)
        print (len(data_json))


# @pytest.fixture(scope="module")
# def client():
#     return myStreamListener()()


def test_fails_without_api_key():
    """An authentication error will be raised."""
    token = None
    myStreamListener = MyStreamListener()
    kwargs = {}
    stream = StreamClient(listener=myStreamListener, token=token, **kwargs)

    pytest.raises(error.AuthenticationError, stream.performs)


def test_fails_with_wrong_fields():
    """An authentication error will be raised."""
    token = None
    myStreamListener = MyStreamListener()
    kwargs = {
        'fields': ['entity']
    }
    stream = StreamClient(listener=myStreamListener, token=token, **kwargs)

    pytest.raises(error.AuthenticationError, stream.performs)



"""Python library for Accern API.

Core client functionality, common across all API requests (including performing
HTTP requests).
"""

try:
    import requests
except ImportError:
    requests = None

from accern import error, http_client, util
from accern.util import six, urlsplit, urlunsplit, urlencode

api_base = "htts://feed.accern.com/v4/stream"


def _api_encode(data):
    for key, value in six.iteritems(data):
        key = util.utf8(key)
        if value is None:
            continue
        elif isinstance(value, list):
            yield (key, ",".join(value))
        else:
            yield (key, util.utf8(value))


class Historical(object):
    """Perform requests to the Accern API web services."""

    def __init__(self, client=None, token=None):
        """Intialize with params.

        :param client: default http client. Optional
        :param token: Accern API token. Required.
        """
        self.api_base = api_base
        self.token = token
        self._client = client or http_client.new_default_http_client()

    def _build_api_headers(self, api_key, method):
        headers = {
            'Authorization': 'Token token=%s' % (api_key)
        }
        return headers

    def _build_api_url(self, url, query):
        scheme, netloc, path, base_query, fragment = urlsplit(url)
        if base_query:
            query = '%s&%s' % (base_query, query)
        return urlunsplit((scheme, netloc, path, query, fragment))

    def request(self, method=None, params=None):
        """Perform HTTP GET/POST with credentials.

        :param url: URL path for the request. Should begin with a slash.

        :param params: HTTP GET parameters.

        :raises ApiError: when the API returns an error.
        :raises Timeout: if the request timed out.
        :raises TransportError: when something went wrong while trying to
            exceute a request.
        """
        if self.token:
            my_token = self.token
        else:
            from accern import token
            my_token = token

        if my_token is None:
            raise error.AuthenticationError('No Token provided')

        abs_url = '%s' % (self.api_base)
        if params is None:
            params = {'token': my_token}
        else:
            params['token'] = my_token
        encoded_params = urlencode(list(_api_encode(params or {})))

        if method == 'get':
            if params:
                abs_url = self._build_api_url(abs_url, encoded_params)
            post_data = None
        elif method == 'post':
            return 'post not implemented yet'
        else:
            raise error.APIConnectionError('Unrecognized HTTP method %r' % (method))

        headers = self._build_api_headers(my_token, method)
        rbody, rcode, rheaders = self._client.request(method, abs_url, headers, post_data)
        return rbody, rcode, rheaders

    def save(self, method, output):
        resp = self.request_raw(method)
        jsonData = util.json.loads(resp.text)
        with open(output, 'w') as outfile:
            util.json.dump(jsonData, outfile)



"""Accern REST APIs.

Core client functionality, common across all API requests (including performing
HTTP requests).
"""

try:
    import requests
except ImportError:
    requests = None

from accern import error, http_client, util
from accern.util import six, urlsplit, urlunsplit, urlencode

api_base = "https://feed.accern.com/v4/alphas"


class API(object):
    """Perform requests to the Accern API web services."""

    def __init__(self, client=None, token=None):
        """Intialize with params.

        :param client: default http client. Optional
        :param token: Accern API token. Required.
        """
        self.api_base = api_base
        self.token = token
        self._client = client or http_client.new_default_http_client()

    def _api_encode(self, data):
        for key, value in six.iteritems(data):
            key = util.utf8(key)
            if value is None:
                continue
            elif isinstance(value, list):
                yield (key, ",".join(value))
            else:
                yield (key, util.utf8(value))

    def _build_api_url(self, url, query):
        scheme, netloc, path, base_query, fragment = urlsplit(url)
        if base_query:
            query = '%s&%s' % (base_query, query)
        return urlunsplit((scheme, netloc, path, query, fragment))

    def filter_resp(self, rbody):
        select = self.select
        data_json = util.json.loads(rbody)['signals']
        data_filtered = []
        for data in data_json:
            if len(select) > 0:
                try:
                    if isinstance(select, list):
                        new_data = dict(zip(select, [data[key] for key in select]))
                    elif isinstance(select, str):
                        new_data = {select: data[select]}
                except KeyError:
                    raise error.AccernError('Invalid fields passed.')
            else:
                new_data = data

            if 'entity_competitors' in new_data:
                new_data['entity_competitors'] = ' | '.join(new_data['entity_competitors'])
            if 'entity_indices' in new_data:
                new_data['entity_indices'] = ' | '.join(new_data['entity_indices'])

            data_filtered.append(new_data)
        resp = {
            'first_id': util.json.loads(rbody)['first_id'],
            'last_id': util.json.loads(rbody)['last_id'],
            'total': util.json.loads(rbody)['total'],
            'signals': data_filtered
        }
        return resp

    def interpret_response(self, rbody, rcode, rheaders):
        try:
            if hasattr(rbody, 'decode'):
                rbody = rbody.decode('utf-8')
            resp = self.filter_resp(rbody)

        except Exception:
            raise error.APIError(
                "Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody, rcode, rheaders)
        if not (200 <= rcode < 300):
            raise error.APIError('API request failed.')

        return resp

    def request(self, method, **kwargs):
        self.select = kwargs.get('select', [])
        self.params = kwargs.get('filter', {})
        rbody, rcode, rheaders = self.request_raw(method.lower())
        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp

    def request_raw(self, method=None):
        """Perform HTTP GET with credentials.

        :param method: HTTP method
        :param params: HTTP GET parameters.
        :raises AuthenticationError: when the token is invalid.
        """
        if self.token:
            my_token = self.token
        else:
            from accern import token
            my_token = token

        if my_token is None:
            raise error.AuthenticationError('No token provided.')

        abs_url = self.api_base
        if self.params is None:
            self.params = {'token': my_token}
        else:
            self.params['token'] = my_token
        encoded_params = urlencode(list(self._api_encode(self.params)))

        if method == 'get':
            if self.params:
                abs_url = self._build_api_url(abs_url, encoded_params)
        else:
            raise error.APIConnectionError('Unsupported HTTP method %r' % (method))

        rbody, rcode, rheaders = self._client.request(method, abs_url, headers=None, post_data=None)
        return rbody, rcode, rheaders



from accern import error
import sys
import textwrap


try:
    import requests
except ImportError:
    requests = None
else:
    try:
        # Require version 0.8.8, but don't want to depend on distutils
        version = requests.__version__
        major, minor, patch = [int(i) for i in version.split('.')]
    except Exception:
        # Probably some new-fangled version, so it should support verify
        pass
    else:
        if (major, minor, patch) < (0, 8, 8):
            sys.stderr.write(
                'Warning: the Accern library requires that your Python '
                '"requests" library be newer than version 0.8.8, but your '
                '"requests" library is version %s. Accern will fall back to '
                'an alternate HTTP library so everything should work. We '
                'recommend upgrading your "requests" library. If you have any '
                'questions, please contact support@accern.com. (HINT: running '
                '"pip install -U requests" should upgrade your requests '
                'library to the latest version.)' % (version,))
            requests = None


__all__ = [
    'new_default_http_client'
]


def new_default_http_client(*args, **kwargs):
    return RequestsClient(*args, **kwargs)


class HTTPClient(object):
    def request(self, method, url, headers, post_data=None):
        raise NotImplementedError(
            'HTTPClient subclasses must implement `request`')


class RequestsClient(HTTPClient):
    name = 'requests'

    def __init__(self, timeout=80, session=None, **kwargs):
        super(RequestsClient, self).__init__(**kwargs)
        self._timeout = timeout
        self._session = session or requests.Session()

    def request(self, method, url, headers, post_data=None):
        kwargs = {}
        try:
            try:
                result = self._session.request(method,
                                               url,
                                               headers=headers,
                                               data=post_data,
                                               timeout=self._timeout,
                                               **kwargs)
            except TypeError as e:
                raise TypeError(
                    'Warning: It looks like your "requests" library is out of '
                    'date. You can fix that by running "pip install -U requests".) '
                    'The underlying error was: %s' % (e,))
            content = result.content
            status_code = result.status_code
        except Exception as e:
            # Would catch just requests.exceptions.RequestException, but can
            # also raise ValueError, RuntimeError, etc.
            self._handle_request_error(e)
        return content, status_code, result.headers

    def _handle_request_error(self, e):
        if isinstance(e, requests.exceptions.RequestException):
            msg = ("Unexpected error communicating with Accern.  "
                   "If this problem persists, let us know at "
                   "support@accern.com.")
            err = "%s: %s" % (type(e).__name__, str(e))
        else:
            msg = ("Unexpected error communicating with Accern. "
                   "It looks like there's probably a configuration "
                   "issue locally.  If this problem persists, let us "
                   "know at support@accern.com.")
            err = "A %s was raised" % (type(e).__name__,)
            if str(e):
                err += " with error message %s" % (str(e),)
            else:
                err += " with no error message"
        msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
        raise error.APIConnectionError(msg)



from accern import API as AccernAPI
from accern import error
import pytest


def test_fails_without_token():
    API = AccernAPI()
    API.token = None
    with pytest.raises(error.AuthenticationError):
        API.request('get')



"""Accern Streaming API client.

Core client functionality, common across all API requests (including performing
HTTP requests).
"""
from accern import error, util
from accern.util import datetime, six, urlsplit, urlunsplit, urlencode
import codecs

import re
import requests
import sys
import time

api_base = "https://feed.accern.com/v4/stream"
end_of_field = re.compile(r'\r\n\r\n|\r\r|\n\n')


class Event(object):
    SSE_LINE_PATTERN = re.compile('(?P<name>[^:]*):?( ?(?P<value>.*))?')

    def __init__(self, data='', event='message', id=None, retry=None):
        self.data = data
        self.event = event
        self.id = id
        self.retry = retry

    def dump(self):
        lines = []
        if self.id:
            lines.append('id: %s' % self.id)

        # Only include an event line if it's not the default already.
        if self.event != 'message':
            lines.append('event: %s' % self.event)

        if self.retry:
            lines.append('retry: %s' % self.retry)

        lines.extend('data: %s' % d for d in self.data.split('\n'))
        return '\n'.join(lines) + '\n\n'

    @classmethod
    def parse(cls, raw):
        """Given a possibly-multiline string representing an SSE message, parse it and return a Event object."""
        msg = cls()
        for line in raw.splitlines():
            m = cls.SSE_LINE_PATTERN.match(line)
            if m is None:
                # Malformed line.  Discard but warn.
                sys.stderr.write('Invalid SSE line: "%s"' % line, SyntaxWarning)
                continue

            name = m.group('name')
            if name == '':
                # line began with a ":", so is a comment.  Ignore
                continue
            value = m.group('value')

            if name == 'data':
                # If we already have some data, then join to it with a newline.
                # Else this is it.
                if msg.data:
                    msg.data = '%s\n%s' % (msg.data, value)
                else:
                    msg.data = value
            elif name == 'event':
                msg.event = value
            elif name == 'id':
                msg.id = value
            elif name == 'retry':
                msg.retry = int(value)

        return msg

    def __str__(self):
        return self.data


class StreamListener(object):
    def on_data(self, raw_data):
        """Call when raw data is received from connection.

        Override this method if you want to manually handle stream data. Return
        False to stop stream and close connection.
        """
        data = util.json.loads(raw_data)
        if 'disconnect' in data:
            if self.on_disconnect(data['disconnect']) is False:
                return False
        return raw_data

    def on_disconnect(self, notice):
        """Call when server return a disconnect notice."""
        return


class StreamClient(object):
    """Perform requests to the Accern API web services."""

    def __init__(self, listener, token, **kwargs):
        """Intialize with params.

        :param client: default http client. Optional
        :param token: Accern API token. Required.
        """
        self.api_base = api_base
        # Keep data here as it streams in
        self.buf = u''
        self.chunk_size = kwargs.get("chunk_size", 1024)
        self.last_id = kwargs.get("last_id", None)
        self._listener = listener or StreamListener()
        self.new_session()
        # Any extra kwargs will be fed into the requests.get call later.
        self.requests_kwargs = kwargs.get("filter", {})
        self.select = kwargs.get("select", None)
        self.retry = kwargs.get("retry", 3000)
        self.timeout = kwargs.get('timeout', 300.0)
        self.token = token

        # The SSE spec requires making requests with Cache-Control: nocache
        if 'headers' not in self.requests_kwargs:
            self.requests_kwargs['headers'] = {}
        self.requests_kwargs['headers']['Cache-Control'] = 'no-cache'

        # The 'Accept' header is not required, but explicit > implicit
        self.requests_kwargs['headers']['Accept'] = 'text/event-stream'

    def __iter__(self):
        return self

    def __next__(self):
        decoder = codecs.getincrementaldecoder(self.resp.encoding)(errors='replace')
        while not self._event_complete():
            try:
                next_chunk = next(self.resp_iterator)
                if not next_chunk:
                    raise EOFError()
                self.buf += decoder.decode(next_chunk)

            except (StopIteration, requests.RequestException, EOFError):
                time.sleep(self.retry / 1000.0)
                self._run()

                # The SSE spec only supports resuming from a whole message, so
                # if we have half a message we should throw it out.
                head, sep, tail = self.buf.rpartition('\n')
                self.buf = head + sep
                continue

        # Split the complete event (up to the end_of_field) into event_string,
        # and retain anything after the current complete event in self.buf
        # for next time.
        (event_string, self.buf) = re.split(end_of_field, self.buf, maxsplit=1)
        msg = Event.parse(event_string)

        # If the server requests a specific retry delay, we need to honor it.
        if msg.retry:
            self.retry = msg.retry

        # last_id should only be set if included in the message.  It's not
        # forgotten if a message omits it.
        if msg.id:
            self.last_id = msg.id

        if msg.data:
            data = self.filter_data(msg.data)
            if self._listener.on_data(data) is False:
                return False

        return msg

    def _api_encode(self, data):
        for key, value in six.iteritems(data):
            key = util.utf8(key)
            if value is None:
                continue
            elif isinstance(value, list):
                yield (key, ",".join(value))
            else:
                yield (key, util.utf8(value))

    def _build_api_url(self, url, query):
        scheme, netloc, path, base_query, fragment = urlsplit(url)
        if base_query:
            query = '%s&%s' % (base_query, query)
        return urlunsplit((scheme, netloc, path, query, fragment))

    def _event_complete(self):
        return re.search(end_of_field, self.buf) is not None

    def _run(self):
        if self.last_id:
            self.requests_kwargs['headers']['Last-Event-ID'] = self.last_id
        # Use session if set.  Otherwise fall back to requests module.
        requester = self.session or requests
        self.resp = requester.get(self.url, stream=True)
        self.resp_iterator = self.resp.iter_content(chunk_size=self.chunk_size)

        # TODO(redirect): Ensure we're handling redirects.  Might also stick the 'origin'
        # attribute on Events like the Javascript spec requires.
        self.resp.raise_for_status()
        while next(self, None) is not None:
            next(self, None)

    def filter_data(self, data):
        select = self.select
        data_json = util.json.loads(util.json.loads(data)['data'])['signals']
        data_filtered = []
        for data in data_json:
            if select is not None and len(select) > 0:
                try:
                    if isinstance(select, list):
                        new_data = dict(zip(select, [data[key] for key in select]))
                    elif isinstance(select, str):
                        new_data = {select: data[select]}
                except KeyError:
                    raise error.AccernError('Invalid fields passed.')
            else:
                new_data = data

            if 'entity_competitors' in new_data:
                new_data['entity_competitors'] = ' | '.join(new_data['entity_competitors'])
            if 'entity_indices' in new_data:
                new_data['entity_indices'] = ' | '.join(new_data['entity_indices'])

            data_filtered.append(new_data)

        return util.json.dumps(data_filtered)

    def new_session(self):
        self.session = requests.Session()
        self.session.params = None

    def performs(self):
        """Perform HTTP GET/POST with credentials.

        :param output: output config.

        :param params: HTTP GET parameters.

        :raises ApiError: when the API returns an error.
        :raises Timeout: if the request timed out.
        :raises TransportError: when something went wrong while trying to
            exceute a request.
        """
        print ("%s - Start streaming, use [Ctrl+C] to stop..." % (datetime.now()))
        if self.token:
            my_token = self.token
        else:
            from accern import token
            my_token = token

        if my_token is None:
            raise error.AuthenticationError('No token provided.')

        abs_url = '%s' % (self.api_base)

        params = self.requests_kwargs
        if params is None:
            params = {'token': my_token}
        else:
            params['token'] = my_token
        encoded_params = urlencode(list(self._api_encode(params or {})))

        self.url = self._build_api_url(abs_url, encoded_params)
        try:
            self._run()
        except KeyboardInterrupt:
            print ("%s - Streaming stopped..." % (datetime.now()))
        else:
            pass

    if six.PY2:
        next = __next__



"""Accern: python library for Accern API.

Accern is a python library to query, download, filter and save Accern Data.
"""

from accern.stream import StreamClient, StreamListener
from accern.api import API

token = None


__all__ = [
    'StreamClient',
    'StreamListener',
    'API'
]
